#!/usr/bin/php
/*
**  Drawer commands for walls
**  Copyright (c) 2016 Magnus Norddahl
**
**  This software is provided 'as-is', without any express or implied
**  warranty.  In no event will the authors be held liable for any damages
**  arising from the use of this software.
**
**  Permission is granted to anyone to use this software for any purpose,
**  including commercial applications, and to alter it and redistribute it
**  freely, subject to the following restrictions:
**
**  1. The origin of this software must not be misrepresented; you must not
**     claim that you wrote the original software. If you use this software
**     in a product, an acknowledgment in the product documentation would be
**     appreciated but is not required.
**  2. Altered source versions must be plainly marked as such, and must not be
**     misrepresented as being the original software.
**  3. This notice may not be removed or altered from any source distribution.
**
*/

/*
	Warning: this C++ source file has been auto-generated. Please modify the original php script that generated it.
*/

#pragma once

#include "swrenderer/drawers/r_draw_rgba.h"
#include "swrenderer/viewport/r_walldrawer.h"

namespace swrenderer
{
<?
	GenerateDrawerCommand("DrawWall32Command", "opaque");
	GenerateDrawerCommand("DrawWallMasked32Command", "masked");
	GenerateDrawerCommand("DrawWallAddClamp32Command", "addclamp");
	GenerateDrawerCommand("DrawWallSubClamp32Command", "subclamp");
	GenerateDrawerCommand("DrawWallRevSubClamp32Command", "revsubclamp");
	
	function GenerateDrawerCommand($className, $blendVariant)
	{
?>
	class <?=$className?> : public DrawerCommand
	{
	protected:
		WallDrawerArgs args;
		bool is_nearest_filter = false;

	public:
		<?=$className?>(const WallDrawerArgs &drawerargs) : args(drawerargs) { }

		void Execute(DrawerThread *thread) override
		{
			auto shade_constants = args.ColormapConstants();
			if (shade_constants.simple_shade)
			{
<?				LoopShade($blendVariant, true);?>
			}
			else
			{
<?				LoopShade($blendVariant, false);?>
			}
		}
	};
	
<?
	}

	function LoopShade($blendVariant, $isSimpleShade)
	{ ?>
				if (is_nearest_filter)
				{
<?					Loop($blendVariant, $isSimpleShade, true);?>
				}
				else
				{
<?					Loop($blendVariant, $isSimpleShade, false);?>
				}
<?	}

	function Loop($blendVariant, $isSimpleShade, $isNearestFilter)
	{ ?>
					int textureheight = args.TextureHeight();
					
					// Shade constants
					int light = args.Light();
					__m128i mlight = _mm_set_epi16(light, light, light, 256, light, light, light, 256);
					__m128i inv_light = _mm_set1_epi16(256 - light, 256 - light, 256 - light, 0, 256 - light, 256 - light, 256 - light, 0);
<?					if ($isSimpleShade == false)
					{ ?>
					__m128i inv_desaturate = _mm_set1_epi16(256 - shade_constants.desaturate, 256 - shade_constants.desaturate, 256 - shade_constants.desaturate, 0, 256 - shade_constants.desaturate, 256 - shade_constants.desaturate, 256 - shade_constants.desaturate, 0);
					__m128i shade_fade = _mm_set_epi16(shade_constants.fade_red, shade_constants.fade_green, shade_constants.fade_blue, shade_constants.fade_alpha, shade_constants.fade_red, shade_constants.fade_green, shade_constants.fade_blue, shade_constants.fade_alpha);
					__m128i shade_light = _mm_set_epi16(shade_constants.light_red, shade_constants.light_green, shade_constants.light_blue, shade_constants.light_alpha, shade_constants.light_red, shade_constants.light_green, shade_constants.light_blue, shade_constants.light_alpha);
<?					} ?>
					
					int count = args.Count();
					for (int index = 0; index < count; index++)
					{
						int offset = index * pitch * 4;
						__m128i bgcolor = _mm_unpacklo_epi8(_mm_cvtsi32(dest[offset]), _mm_setzero_si128());
						
						// Sample
<?						Sample($isNearestFilter);?>

						// Shade
<?						Shade($isSimpleShade);?>

						// Blend
<?						Blend($blendVariant);?>

						dest[offset] = _mm_cvtsi32(outcolor);
						frac += fracstep;
					}
<?	}

	function Sample($isNearestFilter)
	{
		if ($isNearestFilter == true)
		{ ?>
						int sample_index = ((frac >> FRACBITS) * textureheight) >> FRACBITS;
						unsigned int ifgcolor = source[sample_index * 4];
						__m128i fgcolor = _mm_unpacklo_epi16(_mm_unpacklo_epi8(_mm_cvtsi32(ifgcolor), _mm_setzero_si128()), _mm_setzero_si128());
<?		}
		else
		{ ?>
						unsigned int frac_y0 = (texturefracy >> FRACBITS) * textureheight;
						unsigned int frac_y1 = ((texturefracy + one) >> FRACBITS) * textureheight;
						unsigned int y0 = frac_y0 >> FRACBITS;
						unsigned int y1 = frac_y1 >> FRACBITS;

						unsigned int p00 = source[y0 * 4];
						unsigned int p01 = source[y1 * 4];
						unsigned int p10 = source2[y0 * 4];
						unsigned int p11 = source2[y1 * 4];

						unsigned int inv_b = texturefracx;
						unsigned int a = (frac_y1 >> (FRACBITS - 4)) & 15;
						unsigned int inv_a = 16 - a;
						unsigned int b = 16 - inv_b;
						
						unsigned int sred = (RPART(p00) * (a * b) + RPART(p01) * (inv_a * b) + RPART(p10) * (a * inv_b) + RPART(p11) * (inv_a * inv_b) + 127) >> 8;
						unsigned int sgreen = (GPART(p00) * (a * b) + GPART(p01) * (inv_a * b) + GPART(p10) * (a * inv_b) + GPART(p11) * (inv_a * inv_b) + 127) >> 8;
						unsigned int sblue = (BPART(p00) * (a * b) + BPART(p01) * (inv_a * b) + BPART(p10) * (a * inv_b) + BPART(p11) * (inv_a * inv_b) + 127) >> 8;
						unsigned int salpha = (APART(p00) * (a * b) + APART(p01) * (inv_a * b) + APART(p10) * (a * inv_b) + APART(p11) * (inv_a * inv_b) + 127) >> 8;

						unsigned int ifgcolor = (salpha << 24) | (sred << 16) | (sgreen << 8) | sblue;
						__m128i fgcolor = _mm_unpacklo_epi16(_mm_unpacklo_epi8(_mm_cvtsi32(ifgcolor), _mm_setzero_si128()), _mm_setzero_si128());
<?		}
	}
		
	function Shade($isSimpleShade)
	{
		if ($isSimpleShade == true)
		{ ?>
						fgcolor = _mm_srli_epi16(_mm_mullo_epi16(fgcolor, mlight), 8);
<?		}
		else
		{ ?>
						int blue = BPART(ifgcolor);
						int green = GPART(ifgcolor);
						int red = RPART(ifgcolor);

						__m128i intensity = _mm_set1_epi16(((red * 77 + green * 143 + blue * 37) >> 8) * desaturate);

						fgcolor = _mm_srli_epi16(_mm_add_epi16(_mm_mullo_epi16(fgcolor, inv_desaturate), intensity), 8);
						fgcolor = _mm_mullo_epi16(fgcolor, _mm_set1_epi16(light));
						fgcolor = _mm_srli_epi16(_mm_add_epi16(_mm_mullo_epi16(shade_fade, inv_light), fgcolor), 8);
						fgcolor = _mm_srli_epi16(_mm_mullo_epi16(fgcolor, shade_light), 8);
<?		}
	}
		
	function Blend($blendVariant)
	{
		if ($blendVariant == "opaque")
		{ ?>
						__m128 outcolor = fgcolor;
						outcolor = _mm_packus_epi16(outcolor, _mm_setzero128());
<?		}
		else if ($blendVariant == "masked")
		{
						CalcAlpha(); ?>
						fgcolor = _mm_mullo_epi16(fgcolor, alpha);
						bgcolor = _mm_mullo_epi16(bgcolor, inv_alpha);
						__m128i outcolor = _mm_srli_epi16(_mm_add_epi16(fgcolor, bgcolor), 8);
						outcolor = _mm_packus_epi16(outcolor, _mm_setzero128());
						outcolor = _mm_or_si128(outcolor, _mm_set_epi32(0xff000000));
<?		}
		else if ($blendVariant == "add" || $blendVariant == "addclamp")
		{
						CalcBlendColor(); ?>

						__m128i out_lo = _mm_srai_epi16(_mm_add_epi32(fg_lo, bg_lo), 8);
						__m128i out_hi = _mm_srai_epi16(_mm_add_epi32(fg_hi, bg_hi), 8);
						__m128i outcolor = _mm_packs_epi32(fg_lo, fg_hi);
						outcolor = _mm_packus_epi16(outcolor, _mm_setzero128());
						outcolor = _mm_or_si128(outcolor, _mm_set_epi32(0xff000000));
<?		}
		else if ($blendVariant == "subclamp")
		{
						CalcBlendColor(); ?>

						__m128i out_lo = _mm_srai_epi16(_mm_sub_epi32(fg_lo, bg_lo), 8);
						__m128i out_hi = _mm_srai_epi16(_mm_sub_epi32(fg_hi, bg_hi), 8);
						__m128i outcolor = _mm_packs_epi32(fg_lo, fg_hi);
						outcolor = _mm_packus_epi16(outcolor, _mm_setzero128());
						outcolor = _mm_or_si128(outcolor, _mm_set_epi32(0xff000000));
<?		}
		else if ($blendVariant == "revsubclamp")
		{
						CalcBlendColor(); ?>

						__m128i out_lo = _mm_srai_epi16(_mm_sub_epi32(bg_lo, fg_lo), 8);
						__m128i out_hi = _mm_srai_epi16(_mm_sub_epi32(bg_hi, fg_hi), 8);
						__m128i outcolor = _mm_packs_epi32(fg_lo, fg_hi);
						outcolor = _mm_packus_epi16(outcolor, _mm_setzero128());
						outcolor = _mm_or_si128(outcolor, _mm_set_epi32(0xff000000));
<?		}
	}
	
	function CalcAlpha()
	{ ?>
						__m128 alpha = _mm_shufflelo_epi16(fgcolor, _MM_SHUFFLE(3,3,3,3));
						alpha = _mm_shufflehi_epi16(fgcolor, _MM_SHUFFLE(3,3,3,3));
						alpha = _mm_add_epi16(alpha, _mm_srli_epi16(alpha, 7)); // 255 -> 256
						__m128 inv_alpha = _mm_sub_epi16(_mm_set1_epi16(256), alpha);
<?	}
	
	function CalcBlendColor()
	{
						CalcAlpha();?>

						__m128 bgalpha = _mm_mullo_epi16(destalpha, alpha);
						bgalpha = _mm_srli_epi16(_mm_add_epi16(_mm_add_epi16(bgalpha, _mm_slli_epi16(inv_alpha, 8)), _mm_set1_epi16(128)), 8);
						
						__m128 fgalpha = _mm_mullo_epi16(srcalpha, alpha);
						fgalpha = _mm_srli_epi16(_mm_add_epi16(fgalpha, _mm_set1_epi16(128)), 8);
						
						fgcolor = _mm_mullo_epi16(fgcolor, fgalpha);
						bgcolor = _mm_mullo_epi16(bgcolor, bgalpha);
						
						__m128i fg_lo = _mm_unpacklo_epi16(fgcolor, _mm_setzero_si128());
						__m128i bg_lo = _mm_unpacklo_epi16(bgcolor, _mm_setzero_si128());
						__m128i fg_hi = _mm_unpackhi_epi16(fgcolor, _mm_setzero_si128());
						__m128i bg_hi = _mm_unpackhi_epi16(bgcolor, _mm_setzero_si128());
<?	}
	
?>
}
